/*
  SNESAPU 
  copyright (C)2003-04 Alpha-II Productions
  http://www.alpha-ii.com/

  Ported from x86 assembler to C and implementation by Brad Jorsch

  Person who rewrites this.
  Copyright (C) 2004 sanmaiwashi
  http://toriniku.sourceforge.net/
*/

#include "snes9x.h"
#include "memmap.h"
#include "apu_orig.h"
#include "soundux.h"
#include "snesapu.h"
#include "display.h"
#include <math.h>
#include "resampler.h"

/* For note-triggered SPC dump support */
#include "snapshot.h"

typedef float f32;
typedef double f64;
typedef long double f80;

const static struct
{
  int o;
  uint8 m;
} chs[8] =
{
  {
  0x00, 0x01},
  {
  0x10, 0x02},
  {
  0x20, 0x04},
  {
  0x30, 0x08},
  {
  0x40, 0x10},
  {
  0x50, 0x20},
  {
  0x60, 0x40},
  {
0x70, 0x80}};

#undef M_PIl
#define M_PIl		3.1415926535897932384626433832795029L	/* pi */

#define	SA_SNESINT

//#define APUDSP_DEBUG
#ifdef APUDSP_DEBUG
static const char apudsp_log_name[] = "apudsp.log";
static FILE *apudsp_fs = NULL;
extern FILE *trace;
#endif

/***********************************************************************
  Defines
***********************************************************************/

// Envelope mode masks
#define E_TYPE   0x01		// Type of adj: Constant(1/64 or 1/256) / Exp.(255/256)
#define E_DIR    0x02		// Direction: Decrease / Increase
#define E_DEST   0x04		// Destination: Default(0 or 1) / Other(x/8 or .75)
#define E_ADSR   0x08		// Envelope mode: Gain/ADSR
#define E_IDLE   0x80		// Envelope speed is set to 0

#define E_DEC    0x00		// Linear decrease
#define E_EXP    0x01		// Exponential decrease
#define E_INC    0x02		// Linear increase
#define E_BENT   0x06		// Bent line increase
#define E_REL    0x08		// Release mode
#define E_SUST   0x09		// Sustain mode
#define E_ATT    0x0A		// Attack mode
#define E_DECAY  0x0D		// Decay mode
#define E_DIRECT 0x87		// Direct gain

// Envelope adjustment rates ------------------
#define A_GAIN   (1 << E_SHIFT)	// Amount to adjust envelope values
#define A_LIN    ((128*A_GAIN)/64)	// Linear rate to increase/decrease envelope
#define A_KOF    ((128*A_GAIN)/256)	// Rate to decrease envelope during release
#define A_BENT   ((128*A_GAIN)/256)	// Rate to increase envelope after bend
#define A_NOATT  (64*A_GAIN)	// Rate to increase if attack rate is set to 0ms
#define A_EXP    0		// Rate to decrease envelope exponentially (Not used)

// Envelope destination values ----------------
#define D_MAX    ((128*A_GAIN)-1)	// Maximum envelope value
#define D_ATTACK ((128*A_GAIN)-1)	// Destination of attack rate
#define D_BENT   ((128*A_GAIN*3)/4)	// First destination of bent line
#define D_EXP    ((128*A_GAIN)/8)	// Minimum decay destination value
#define D_MIN    0		// Minimum envelope value

// Others -------------------------------------
#define T_EXP    600		// 1->1/10, eVal*255/256, T_EXP times.

#undef ABS
#define ABS(a) ((a) < 0 ? -(a) : (a))

static inline signed sclip4(signed x)
{
  return (((x & 0xf) ^ 0x8) - 0x8);
}

static inline signed sclip15(signed x)
{
  return (((x & 0x7fff) ^ 0x4000) - 0x4000);
}

static inline signed sclip16(signed x)
{
  return (((x & 0xffff) ^ 0x8000) - 0x8000);
}

static inline signed sclamp15(signed x)
{
  return (x > 0x3fff) ? 0x3fff : (x < -0x4000) ? -0x4000 : x;
}

static inline signed sclamp16(signed x)
{
  return (x > 0x7fff) ? 0x7fff : (x < -0x8000) ? -0x8000 : x;
}

/***********************************************************************
  Variables
***********************************************************************/

static int brrTab[256];
static int16 gaussTab[1024];
static int16 cubicTab[1024];
static int16 sincTab[2048];

#ifdef SA_SNESINT
#include "gausspre.h"
#endif

// Register write handler prototypes
static void RADSR1 (int, uint8);
static void RADSR2 (int, uint8);
static void REDl (int, uint8);
static void REFB (int, uint8);
static void REFBM (int, uint8);
static void REVolL (int, uint8);
static void REVolLM (int, uint8);
static void REVolR (int, uint8);
static void REVolRM (int, uint8);
static void RNOn (int, uint8);
static void RFCI (int, uint8);
static void RFlg (int, uint8);
static void RESA (int, uint8);
static void RGain (int, uint8);
static void RKOn (int, uint8);
static void RKOf (int, uint8);
static void RDir (int, uint8);
static void RMVolL (int, uint8);
static void RMVolLM (int, uint8);
static void RMVolR (int, uint8);
static void RMVolRM (int, uint8);
static void RNull (int, uint8);
static void RPMOn (int, uint8);
static void RPitch (int, uint8);
static void RSrcn (int, uint8);
static void RVolL (int, uint8);
static void RVolLM (int, uint8);
static void RVolR (int, uint8);
static void RVolRM (int, uint8);
static void REndX (int, uint8);

// Jump table for DSP register writes (see DSPIn)
static void (*dspRegs[0x80]) (int, uint8) =
{
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, RMVolL, REFB, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, RMVolR, RNull, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, REVolL, RPMOn, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, REVolR, RNOn, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, RKOn, RNull, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, RKOf, RDir, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, RFlg, RESA, RNull, RFCI,
  RVolL, RVolR, RPitch, RPitch, RSrcn, RADSR1, RADSR2, RGain,
  RNull, RNull, RNull, RNull, REndX, REDl, RNull, RFCI
};

// Registers that require special handling for output type
static void (*dspRegsI[8]) (int, uint8) =
{				// Integer
RVolL, RVolR, RMVolL, RMVolR, REVolL, REVolR, REFB, RFCI};

static void (*dspRegsM[8]) (int, uint8) =
{				// Monaural
RVolLM, RVolRM, RMVolLM, RMVolRM, REVolLM, REVolRM, REFBM, RFCI};

// Frequency table -------------------------
const static uint32 freqTab[32] = {
  0,
  2048, 1536, 1280,		// Number of samples between updates.  Used to determine
  1024, 768, 640,		// envelope rates and noise frequencies
  512, 384, 320,
  256, 192, 160,
  128, 96, 80,
  64, 48, 40,
  32, 24, 20,
  16, 12, 10,
  8, 6, 5,
  4, 3,
  2,
  1
};

const static uint16 rate_table[32] = {
  0x0000, 0x000F, 0x0014, 0x0018, 0x001E, 0x0028, 0x0030, 0x003C,
  0x0050, 0x0060, 0x0078, 0x00A0, 0x00C0, 0x00F0, 0x0140, 0x0180,
  0x01E0, 0x0280, 0x0300, 0x03C0, 0x0500, 0x0600, 0x0780, 0x0A00,
  0x0C00, 0x0F00, 0x1400, 0x1800, 0x1E00, 0x2800, 0x3C00, 0x7800
};

const int16 gaussian_table[512] = {
     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
     1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    2,    2,    2,    2,    2,
     2,    2,    3,    3,    3,    3,    3,    4,    4,    4,    4,    4,    5,    5,    5,    5,
     6,    6,    6,    6,    7,    7,    7,    8,    8,    8,    9,    9,    9,   10,   10,   10,
    11,   11,   11,   12,   12,   13,   13,   14,   14,   15,   15,   15,   16,   16,   17,   17,
    18,   19,   19,   20,   20,   21,   21,   22,   23,   23,   24,   24,   25,   26,   27,   27,
    28,   29,   29,   30,   31,   32,   32,   33,   34,   35,   36,   36,   37,   38,   39,   40,
    41,   42,   43,   44,   45,   46,   47,   48,   49,   50,   51,   52,   53,   54,   55,   56,
    58,   59,   60,   61,   62,   64,   65,   66,   67,   69,   70,   71,   73,   74,   76,   77,
    78,   80,   81,   83,   84,   86,   87,   89,   90,   92,   94,   95,   97,   99,  100,  102,
   104,  106,  107,  109,  111,  113,  115,  117,  118,  120,  122,  124,  126,  128,  130,  132,
   134,  137,  139,  141,  143,  145,  147,  150,  152,  154,  156,  159,  161,  163,  166,  168,
   171,  173,  175,  178,  180,  183,  186,  188,  191,  193,  196,  199,  201,  204,  207,  210,
   212,  215,  218,  221,  224,  227,  230,  233,  236,  239,  242,  245,  248,  251,  254,  257,
   260,  263,  267,  270,  273,  276,  280,  283,  286,  290,  293,  297,  300,  304,  307,  311,
   314,  318,  321,  325,  328,  332,  336,  339,  343,  347,  351,  354,  358,  362,  366,  370,
   374,  378,  381,  385,  389,  393,  397,  401,  405,  410,  414,  418,  422,  426,  430,  434,
   439,  443,  447,  451,  456,  460,  464,  469,  473,  477,  482,  486,  491,  495,  499,  504,
   508,  513,  517,  522,  527,  531,  536,  540,  545,  550,  554,  559,  563,  568,  573,  577,
   582,  587,  592,  596,  601,  606,  611,  615,  620,  625,  630,  635,  640,  644,  649,  654,
   659,  664,  669,  674,  678,  683,  688,  693,  698,  703,  708,  713,  718,  723,  728,  732,
   737,  742,  747,  752,  757,  762,  767,  772,  777,  782,  787,  792,  797,  802,  806,  811,
   816,  821,  826,  831,  836,  841,  846,  851,  855,  860,  865,  870,  875,  880,  884,  889,
   894,  899,  904,  908,  913,  918,  923,  927,  932,  937,  941,  946,  951,  955,  960,  965,
   969,  974,  978,  983,  988,  992,  997, 1001, 1005, 1010, 1014, 1019, 1023, 1027, 1032, 1036,
  1040, 1045, 1049, 1053, 1057, 1061, 1066, 1070, 1074, 1078, 1082, 1086, 1090, 1094, 1098, 1102,
  1106, 1109, 1113, 1117, 1121, 1125, 1128, 1132, 1136, 1139, 1143, 1146, 1150, 1153, 1157, 1160,
  1164, 1167, 1170, 1174, 1177, 1180, 1183, 1186, 1190, 1193, 1196, 1199, 1202, 1205, 1207, 1210,
  1213, 1216, 1219, 1221, 1224, 1227, 1229, 1232, 1234, 1237, 1239, 1241, 1244, 1246, 1248, 1251,
  1253, 1255, 1257, 1259, 1261, 1263, 1265, 1267, 1269, 1270, 1272, 1274, 1275, 1277, 1279, 1280,
  1282, 1283, 1284, 1286, 1287, 1288, 1290, 1291, 1292, 1293, 1294, 1295, 1296, 1297, 1297, 1298,
  1299, 1300, 1300, 1301, 1302, 1302, 1303, 1303, 1303, 1304, 1304, 1304, 1304, 1304, 1305, 1305,
};

// -----------------------------------------
static uint8 *pAPURAM;		// -> APU RAM
static uint8 src_buffer[9];	// Temporary

// Mixing ----------------------------------
VoiceMix mix[8];		// Mixing settings for each voice and waveform playback
static uint8 voiceKon;		// Voices that are currently being key on, == APU.KeyedChannels
static uint8 voiceKoff;		//
static uint32 rateTab[32];	// Update Rate Table
static Resampler *resampler = NULL;
static int lag;

// DSP Options ----------------------
static int8 dspChn;		// Number of channels being output
static int8 dspSize;		// Size of samples in bytes
static int32 dspRate;		// Output sample rate
static int32 realRate;		// Real output sample rate, not emulated rate
static uint32 pitchAdj;		// Amount to adjust pitch rates [16.16]
static uint8 dspOpts;		// Option flags passed to SetDSPOpt

// Prototypes for functions that will go in dspDecmp
static inline void UnpckSrc (uint8, uint16 &, int16 *, int32 &, int32 &);

// Volume ----------------------------------
static int32 volAdj;		// Amount to adjust main volumes [-15.16]
static int32 volMainL;		// Main volume
static int32 volMainR;
static int32 volEchoL;		// Echo volume
static int32 volEchoR;

// Echo ------------------------------------
static uint32 echoStart;	// Echo Start Address
static uint32 echoDel;		// Size of delay (in bytes)
static uint32 echoCur;		// Current sample in echo area
static int32 echoFB;		// Echo feedback

// Noise -----------------------------------
static int16 noise_ctr;
static int16 noise_rate;
static int noise;

// Echo filtering --------------------------
uint8 firCur;		// Index of the first sample to feed into the filter
static uint8 disEcho;		// 0 if echo is enabled

// Storage buffers -------------------------
int16 Loop [FIRBUF];
int8 FilterTaps [8];
static uint8 Echo[0x10000];//for SPCPlay

// Other
static uint32 dsp_counter;
static SSoundData *sd;
static Channel *channel[NUM_CHANNELS];

//Used for converting envelope mode flags to Snes9x's enum
static int SAtoS9xEMode[256];

/***********************************************************************
  Code
***********************************************************************/

static inline uint8
DSP_GET_BYTE (uint16 addr)
{
  return pAPURAM[addr];
}

static inline uint16
DSP_GET_WORD (uint16 addr)
{
  return (DSP_GET_BYTE (addr) | (DSP_GET_BYTE (addr + 1) << 8));
}

static inline void
DSP_SET_BYTE (uint16 addr, uint8 data)
{
  if(addr >= 0xf0 && addr <= 0xff)
    return;
  pAPURAM[addr] = data;
}

static inline void
DSP_SET_WORD (uint16 addr, uint16 data)
{
  DSP_SET_BYTE (addr, data);
  DSP_SET_BYTE (addr + 1, data >> 8);
}

static inline uint16
DSP_GET_ECHO (uint16 addr)//for SPCPlay
{
  return (Echo[addr] | (Echo[(addr + 1) & 0xFFFF] << 8));
}

static inline void
DSP_SET_ECHO (uint16 addr, uint16 data)//for SPCPlay
{
  Echo[addr] = data;
  Echo[(addr + 1) & 0xFFFF] = data >> 8;
}

static inline uint16
DSP_GET_SRC (int i)
{
  return DSP_GET_WORD ((APU.DSP[APU_DIR] << 8) +
		       (APU.DSP[chs[i].o + APU_SRCN] << 2));
}

static inline uint16
DSP_GET_SRC_LOOP (int i)
{
  return DSP_GET_WORD ((APU.DSP[APU_DIR] << 8) +
		       (APU.DSP[chs[i].o + APU_SRCN] << 2) + 2);
}

static inline uint8 *
DSP_GET_SRC_P (uint16 addr)
{
  if (addr <= (0x10000 - 9))
    return &pAPURAM[addr];

  int len = 0x10000 - addr;
  memcpy (src_buffer, &pAPURAM[addr], (size_t) len);
  memcpy (src_buffer + len, pAPURAM, (size_t) (9 - len));
  return src_buffer;
}

static inline uint16
DSP_GET_PITCH (int i)
{
  return ((APU.DSP[chs[i].o + APU_P_HIGH] << 8) |
	  APU.DSP[chs[i].o + APU_P_LOW]) & FREQUENCY_MASK;
}

// Start Waveform Decompression
//  Called when a voice is keyed on to set up the internal data for waveform
//  mixing, and decompress the first block.
static inline void
StartSrc (int ch)
{
  mix[ch].mFlg &= ~MFLG_END;

  // Initialize interpolation
  memset (&mix[ch].sBuf[0x10], 0, sizeof (int16) * 16);

  mix[ch].bCur = DSP_GET_SRC (ch);
  mix[ch].sIdx = -9;
  mix[ch].sP1 = 0;
  mix[ch].sP2 = 0;
}

static inline void ChgSilence(int ch);

static inline void
UpdateSrc (int ch, bool8 firstblk)
{
  if(!firstblk)
    {
      if((mix[ch].bHdr & 1))
	{
	  // Loop?
	  if(mix[ch].bHdr & 2)
	    APU.DSP[APU_ENDX] |= chs[ch].m;
	  mix[ch].bCur = DSP_GET_SRC_LOOP (ch);
	}
    }

  mix[ch].bHdr = DSP_GET_BYTE (mix[ch].bCur);

  // Is this the end block?
  if((mix[ch].bHdr & 1) && !(mix[ch].bHdr & 2))
    {
      APU.DSP[APU_ENDX] |= chs[ch].m;
      mix[ch].mFlg |= MFLG_END;
      ChgSilence(ch);
    }

  if(!(mix[ch].mFlg & MFLG_OFF))
    UnpckSrc (mix[ch].bHdr,
	      mix[ch].bCur,
	      &mix[ch].sBuf[mix[ch].sIdx & 0x10],
	      mix[ch].sP1,
	      mix[ch].sP2);
  else
    mix[ch].bCur += 9;
}

static inline void
SetNoiseHertz ()
{
  noise_rate = rate_table[APU.DSP[APU_FLG] & 0x1f];
}

//#define APU_VOL_LEFT 0x00
//#define APU_VOL_RIGHT 0x01
//#define APU_P_LOW 0x02
//#define APU_P_HIGH 0x03
//#define APU_SRCN 0x04
//#define APU_ADSR1 0x05
//#define APU_ADSR2 0x06
//#define APU_GAIN 0x07
static inline void
InitReg (int reg, uint8 val)
{
  APU.DSP[reg] = val;
  int i = ((reg) & 0x70) >> 4;
  /* if(~reg & 0x08) the register is in dsp.voice */
  if (!((~(reg)) & MFLG_OFF & mix[i].mFlg))
    (*dspRegs[reg]) (i, val);
}

bool8
InitAPUDSP ()
{
  int i, c;

  pAPURAM = IAPU.RAM;		// Save pointer to SPC RAM

  sd = &SoundData;
  for (i = 0; i < NUM_CHANNELS; i++)
    {
      channel[i] = &sd->channels[i];
    }

  // Reset values so SetDSPOpt will create new ones
  dspChn = -1;
  dspSize = -1;
  dspRate = -1;
  realRate = SNES_SAMPLE_RATE;
  dspOpts = 0;
  disEcho = 0;

  // Erase all mixer settings
  for (i = 0; i < 8; i++)
    memset (&mix[i], 0, sizeof (VoiceMix));

  // Set pointers to reasonable default values. Otherwise savestates cause
  // SEGVs!
  for (i = 0; i < 8; i++)
    {
      mix[i].sIdx = 0;
      mix[i].bCur = DSP_GET_SRC (i);
    }

  // Build a look-up table for all possible expanded values in a BRR block.
  // range 0-12
  for (i = 0; i < 13; i++)
    {
      for (c = 0; c < 16; c++)
	brrTab[(i << 4) + c] = (int)((int16)(c << 12) >> 12) << i >> 1;
    }
  // range 13-15
  for (i = 13; i < 16; i++)
    {
      for (c = 0; c < 16; c++)
	brrTab[(i << 4) + c] = (int)((int16)(c << 12) >> 12) & ~0x7ff;
    }

#ifdef SA_SNESINT
  //Interleave Gaussian table ---------------
  for (i = 0; i < 256; i++)
    {
      gaussTab[(i << 2) + 3] = gaussPre[i + 0];
      gaussTab[(i << 2) + 2] = gaussPre[i + 256];
      gaussTab[(i << 2) + 1] = gaussPre[i + 512];
      gaussTab[(i << 2) + 0] = gaussPre[i + 768];
    }
#else
  //Build a look-up table based on the Gaussian curve.
  //The table is built from the following equation:
  //
  //                  2
  //            (x/pi)
  //          - -------
  // y = A * e     2
  //
  //       ____
  //     \| 2pi * o
  // A = ----------
  //         n
  //
  //o is the mean value (2^15)
  //n is the number of points to interpolate between (4)
  f80 _512pi = (f80) 512.0 / M_PIl;
  f32 fpA = (f32) ((sqrt (M_PIl * 2) * 32768) / 4);

  for (i = 0; i < 256; i++)
    {
      f80 x;
      x = (f80) ((-512 + i) + 0) / _512pi;
      gaussTab[(i << 2) + 3] = (int16) (exp (-(x * x * 0.5)) * fpA);

      x = (f80) ((-512 + i) + 256) / _512pi;
      gaussTab[(i << 2) + 2] = (int16) (exp (-(x * x * 0.5)) * fpA);

      x = (f80) ((-512 + i) + 512) / _512pi;
      gaussTab[(i << 2) + 1] = (int16) (exp (-(x * x * 0.5)) * fpA);

      x = (f80) ((-512 + i) + 768) / _512pi;
      gaussTab[(i << 2) + 0] = (int16) (exp (-(x * x * 0.5)) * fpA);
    }
#endif

  //Build a look-up table to calculate a cubic spline with only four integer multiplies.
  //The table is built from the following equation, simplified for s:
  //
  // y = ax^3 + bx^2 + cx + d
  //
  //     3 (s[0] - s[1]) - s[-1] + s[2]
  // a = ------------------------------
  //                   2
  //
  //                      5 s[0] + s[2]
  // b = 2 s[1] + s[-1] - -------------
  //                            2
  //
  //     s[1] - s[-1]
  // c = ------------
  //          2
  //
  // d = s[0]
  //
  //y is the return sample
  //x is the delta from current sample
  //s is a four sample array with [0] being the current sample
  f80 x1, x2, x3;
  for (i = 0; i < 256; i++)
    {
      //x1=(n/256)  x2=(n/256)^2  x3=(n/256)^3
      x1 = (f80) i / (f80) 256.0;
      x2 = x1 * x1;
      x3 = x2 * x1;

      //s[-1] *= -.5(x^3) + (x^2) - .5x ------
      cubicTab[(i << 2) + 0] =
	(int16) (((f80) - 0.5 * x1 + x2 + (f80) - 0.5 * x3) * 32767.0);

      //s[0] *= 1.5(x^3) - 2.5(x^2) + 1 ------
      cubicTab[(i << 2) + 1] =
	(int16) (((f80) - 2.5 * x2 + (f80) 1.5 * x3 + 1.0) * 32767.0);

      //s[1] *= -1.5(x^3) + 2(x^2) + .5x -----
      cubicTab[(i << 2) + 2] =
	(int16) (((f80) 0.5 * x1 + x2 + x2 + (f80) - 1.5 * x3) * 32767.0);

      //s[2] *= .5(x^3) - .5(x^2) ------------
      cubicTab[(i << 2) + 3] =
	(int16) (((f80) - 0.5 * x2 + (f80) 0.5 * x3) * 32767.0);
    }

  //Build a look-up table for 8-point sinc interpolation with a Hann window
  //
  //  sin((n/2)pi x)
  //  -------------- * (0.5 + 0.5cos(pi x))
  //    (n/2)pi x
  //
  //n is the number of points of interpolation (8)

  //Set first row to 0 0 0 1 0 0 0 0
  sincTab[0] = sincTab[1] = sincTab[2] = sincTab[3] = 0;
  sincTab[4] = sincTab[5] = sincTab[6] = sincTab[7] = 0;
  sincTab[3] = 32767;

  int32 a = -769;
  for(i = 8; i < 2048; a -= 0x801, i += 8)
    {
      for(c = 0; c < 8; a += 256, c++)
	{
	  f80 x = (f80)a * ((f80)1.0/256.0) * M_PIl;
	  x = (f80)sin(x) / x;

	  f80 w = (f80)a * ((f80)1.0/1024.0) * M_PIl;
	  w = ((f80)1.0 + cos(w)) * 0.5;

	  sincTab[i + c] = (int16)(x * w * 32768.0);
	}
    }

  // Create SAtoS9xEMode table
  for (i = 0; i < 256; i++)
    SAtoS9xEMode[i] = SOUND_SILENT;
  SAtoS9xEMode[E_IDLE] = SOUND_SILENT;
  SAtoS9xEMode[E_ATT] = SOUND_ATTACK;
  SAtoS9xEMode[E_DECAY] = SOUND_DECAY;
  SAtoS9xEMode[E_SUST] = SOUND_SUSTAIN;
  SAtoS9xEMode[E_REL] = SOUND_RELEASE;
  SAtoS9xEMode[E_DIRECT] = SOUND_GAIN;
  SAtoS9xEMode[E_INC] = SOUND_INCREASE_LINEAR;
  SAtoS9xEMode[E_BENT] = SOUND_INCREASE_BENT_LINE;
  SAtoS9xEMode[E_DEC] = SOUND_DECREASE_LINEAR;
  SAtoS9xEMode[E_EXP] = SOUND_DECREASE_EXPONENTIAL;

  SAtoS9xEMode[E_ATT | E_IDLE] = SOUND_GAIN;
  SAtoS9xEMode[E_DECAY | E_IDLE] = SOUND_GAIN;
  SAtoS9xEMode[E_SUST | E_IDLE] = SOUND_GAIN;

  SAtoS9xEMode[E_DEC | E_IDLE] = SOUND_GAIN;
  SAtoS9xEMode[E_EXP | E_IDLE] = SOUND_GAIN;
  SAtoS9xEMode[E_INC | E_IDLE] = SOUND_GAIN;
  SAtoS9xEMode[E_BENT | E_IDLE] = SOUND_GAIN;

  SAtoS9xEMode[E_REL | E_IDLE] = SOUND_GAIN;

  if(!resampler)
    {
      resampler = new Resampler(SYNCBUF*2);
      if(!resampler)
	return FALSE;
    }

  SetAPUDSPOpt (2, 16, SNES_SAMPLE_RATE, 0);

  return TRUE;
}

void DeinitAPUDSP ()
{
  if (resampler)
    {
      delete resampler;
      resampler = NULL;
    }
}

void
SetAPUDSPOpt (int numChn, int bits, int32 rate, int opts)
{
  int i;
  bool8 fixVol = FALSE;

  //=========================================
  // Verify parameters

  // numChn
  if (numChn == -1)
    numChn = dspChn;
  else
    numChn = numChn != 1 ? 2 : 1;

  // bits
  if (bits == -1)
    bits = dspSize;
  else
    {
      switch (bits)
	{
	case 8:
	  break;
	case 16:
	  break;
	case 24:
	  break;
	case 32:
	  break;
	case -32:
	  break;
	default:
	  bits = 16;
	  break;
	}
      bits >>= 3;
    }

  // rate
  if (rate == -1)
    rate = (dspOpts & DSP_ANALOG) ? realRate : dspRate;
  else
    rate = (rate >= 8000 && rate <= 192000) ? rate : SNES_SAMPLE_RATE;

  // opts
  if (opts == -1)
    opts = dspOpts;

  // Reverse stereo
  if ((dspOpts & DSP_REVERSE) ^ (opts & DSP_REVERSE))	// Did reverse flag change?
    fixVol = TRUE;

  // Disable echo ----------------------------
  disEcho = (disEcho & ~DSP_NOECHO) | (opts & DSP_NOECHO);

  // Save option flags
  dspOpts = (uint8) opts;

  SetAPUDSPAmp (FIXED_POINT * Options.VolumeLevel);

  //=========================================
  // Calculate sample rate change
  realRate = rate;
  dspOpts = rate < SNES_SAMPLE_RATE ? (dspOpts & ~DSP_ANALOG) : dspOpts;

  // Has sample rate changed?
  if (rate != dspRate)
    {
      dspRate = rate;

      pitchAdj = FIXED_POINT << 4;

      uint64 r, rr;

      r = (uint64) 0x1000 *pitchAdj;
      rateTab[0] = 0;
      for (i = 1; i < 32; i++)
	{
	  rr = r / freqTab[i];
	  rateTab[i] =
	    (uint32) ((rr >> FIXED_POINT_SHIFT) +
		      ((rr & FIXED_POINT_REMAINDER) ? 1 : 0));
	}

      // Adjust voice rates
      for (i = 0; i < 7; i++)
	{
	  mix[i].mOrgP = DSP_GET_PITCH (i);
	  r = (uint64) mix[i].mOrgP * pitchAdj;
	  mix[i].mOrgRate =
	    (uint32) ((r >> FIXED_POINT_SHIFT) +
		      ((r & FIXED_POINT_REMAINDER) ? 1 : 0));
	  mix[i].mRate = mix[i].mOrgRate;
	  mix[i].mDec = 0;
	  mix[i].eRate = rateTab[mix[i].eRIdx];
	  mix[i].eDec = 0;
	}

      double time_ratio = (double)RATE_64K / (dspRate * 2);
      resampler->time_ratio(time_ratio);
    }

  //=========================================
  // Set sample size
  if (bits != dspSize)
    dspSize = bits;

  //=========================================
  // Set number of channels
  if (numChn != dspChn)
    {
      dspChn = numChn;
      fixVol = TRUE;
    }

  //=========================================
  // Fixup volume handlers
  if (fixVol)
    {
      // Setup registers
      void (**regs) (int, uint8);
      regs = dspChn == 1 ? dspRegsM : dspRegsI;

      if (!(dspOpts & DSP_REVERSE))
	{
	  for (i = 0; i < 8; i++)
	    {
	      dspRegs[chs[i].o + APU_VOL_LEFT] = regs[0];	// VOLL handler
	      dspRegs[chs[i].o + APU_VOL_RIGHT] = regs[1];	// VOLR handler
	    }
	  dspRegs[APU_MVOL_LEFT] = regs[2];	// MVOLL handler
	  dspRegs[APU_MVOL_RIGHT] = regs[3];	// MVOLR handler
	  dspRegs[APU_EVOL_LEFT] = regs[4];	// EVOLL handler
	  dspRegs[APU_EVOL_RIGHT] = regs[5];	// EVOLR handler
	}
      else
	{
	  // Reverse, e.g. put VOLL in VOLR slot
	  for (i = 0; i < 8; i++)
	    {
	      dspRegs[chs[i].o + APU_VOL_LEFT] = regs[1];	// VOLR handler
	      dspRegs[chs[i].o + APU_VOL_RIGHT] = regs[0];	// VOLL handler
	    }
	  dspRegs[APU_MVOL_LEFT] = regs[3];	// MVOLR handler
	  dspRegs[APU_MVOL_RIGHT] = regs[2];	// MVOLL handler
	  dspRegs[APU_EVOL_LEFT] = regs[5];	// EVOLR handler
	  dspRegs[APU_EVOL_RIGHT] = regs[4];	// EVOLL handler
	}
      for (i = 0; i < 8; i++)
	dspRegs[chs[i].o + APU_C0] = regs[7];	// FC handler
      dspRegs[APU_EFB] = regs[6];	// EFB handler

      // Fix volumes and filter
      for (i = 0; i < 8; i++)
	{
	  InitReg (chs[i].o + APU_VOL_LEFT, APU.DSP[chs[i].o + APU_VOL_LEFT]);
	  InitReg (chs[i].o + APU_VOL_RIGHT,
		   APU.DSP[chs[i].o + APU_VOL_RIGHT]);
	  InitReg (chs[i].o + APU_C0, APU.DSP[chs[i].o + APU_C0]);
	}
      InitReg (APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
      InitReg (APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
      InitReg (APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
      InitReg (APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

      // Fix feedback level
      InitReg (APU_EFB, APU.DSP[APU_EFB]);
    }
}

void
ResetAPUDSP (void)
{
  int i;

  for (i = 0; i < 8; i++)
    {
      // Voice is inactive
      mix[i].mFlg &= (MFLG_MUTE | MFLG_OFF);
      mix[i].mFlg |= MFLG_OFF;
    }

  // Disable echo
  disEcho &= DSP_NOECHO;

  // Disable echo if all taps are 0
  disEcho |= 1;

  // Erase DSP Registers ---------------------
  memset (APU.DSP, 0, 0x80);

  APU.DSP[APU_KON] = 0;
  APU.DSP[APU_KOFF] = 0;

  // Place DSP in power up mode
  APU.DSP[APU_FLG] = APU_SOFT_RESET | APU_MUTE | APU_ECHO_DISABLED;

  APU.KeyedChannels = 0;

  dsp_counter = 0;

  // Erase internal mixing settings
  // We'll just save-and-restore mFlg instead of selectively zeroing
  for (i = 0; i < 8; i++)
    {

      // Set pointers to reasonable default values. Otherwise savestates
      // cause SEGVs!
      mix[i].bCur = DSP_GET_SRC (i);
      mix[i].bHdr = DSP_GET_BYTE (mix[i].bCur);

      // Envelope
      ChgSilence(i);
      mix[i].eDec = 0;

      // Samples
      memset (mix[i].sBuf, 0, sizeof (int16) * 32);
      mix[i].sIdx = 0;
      mix[i].sP1 = 0;
      mix[i].sP2 = 0;

      // Mixing
      mix[i].mChnL = 0;
      mix[i].mChnR = 0;
      RPitch (i, APU.DSP[chs[i].o + APU_P_LOW]);
      mix[i].mDec = 0;
      mix[i].mOut = 0;
    }

  // Erase global volume settings
  volMainL = 0;
  volMainR = 0;
  volEchoL = 0;
  volEchoR = 0;

  // Erase noise settings
  noise_ctr = 0;
  noise_rate = 0;
  noise = 0x4000;

  // Erase echo region
  memset (Echo, 0, 0x10000);

  // Reset echo variables
  echoStart = 0;
  echoCur = 0;
  echoDel = 2;			// Delay 1 sample
  echoFB = 0;

  // Echo filter
  memset (Loop, 0, sizeof (int16) * FIRBUF);	// Erase filter memory
  memset (FilterTaps, 0, sizeof (int8) * 8);	// Reset filter coefficients

  // Reset filter variables
  firCur = 0;

  // Disable voices
  voiceKon = 0;
  voiceKoff = 0;
}

void
StoreAPUDSP ()
{
  int i;

  APU.KeyedChannels = voiceKon;

  sd->master_volume_left = (short) (int8) APU.DSP[APU_MVOL_LEFT];
  sd->master_volume_right = (short) (int8) APU.DSP[APU_MVOL_RIGHT];
  sd->echo_volume_left = (short) (int8) APU.DSP[APU_EVOL_LEFT];
  sd->echo_volume_right = (short) (int8) APU.DSP[APU_EVOL_RIGHT];
  sd->echo_enable = APU.DSP[APU_EON];
  sd->echo_feedback = (int) (int8) APU.DSP[APU_EFB];

  sd->echo_buffer_size =
    (512 * (APU.DSP[APU_EDL] & 0x0f) * SNES_SAMPLE_RATE) / SNES_SAMPLE_RATE;
  sd->echo_buffer_size <<= 1;

  sd->echo_ptr = 0;

  sd->echo_write_enabled = (APU.DSP[APU_FLG] & APU_ECHO_DISABLED) == 0;
  sd->echo_channel_enable = APU.DSP[APU_EON];
  sd->pitch_mod = APU.DSP[APU_PMON];

  sd->dummy[0] = 0;
  sd->dummy[1] = 0;
  sd->dummy[2] = 0;

  //no_filter
  S9xSetFilterCoefficient (0, (int) (signed char) APU.DSP[APU_C0]);

  sd->master_volume[0] = sd->master_volume_left;
  sd->master_volume[1] = sd->master_volume_right;
  sd->echo_volume[0] = sd->echo_volume_left;
  sd->echo_volume[1] = sd->echo_volume_right;

  uint32 freq, f;

  freq = freqTab[APU.DSP[APU_FLG] & 0x1f];
  if (freq > 0)
    {
      f = (SNES_SAMPLE_RATE << FIXED_POINT_SHIFT) / freq;
      sd->noise_hertz =
	(int) ((f >> FIXED_POINT_SHIFT) +
	       ((f & FIXED_POINT_REMAINDER) ? 1 : 0));
    }
  else
    sd->noise_hertz = 0;

  for (i = 0; i < 8; i++)
    {
      int32 eVal;

      if (mix[i].mFlg & MFLG_OFF)
	{			// Channel is inactive
	  channel[i]->state = SOUND_SILENT;
	  channel[i]->mode = MODE_NONE;
	}
      else
	{
	  channel[i]->state = SAtoS9xEMode[mix[i].eMode];
	  if (mix[i].mFlg & MFLG_KOFF)	// Key off state
	    channel[i]->mode = MODE_RELEASE;
	  else
	    {
	      if (APU.DSP[chs[i].o + APU_ADSR1] & 0x80)	// ADSR mode
		channel[i]->mode = MODE_ADSR;
	      else
		{		// GAIN mode
		  if (!(APU.DSP[chs[i].o + APU_GAIN] & 0x80))	// Direct mode
		    channel[i]->mode = MODE_GAIN;
		  else
		    {		// Indirect mode
		      switch (APU.DSP[chs[i].o + APU_GAIN] & 0x60)
			{
			case 0x00:
			  channel[i]->mode = MODE_DECREASE_LINEAR;
			  break;

			case 0x20:
			  channel[i]->mode = MODE_DECREASE_EXPONENTIAL;
			  break;

			case 0x40:
			  channel[i]->mode = MODE_INCREASE_LINEAR;
			  break;

			case 0x60:
			  channel[i]->mode = MODE_INCREASE_BENT_LINE;
			  break;
			}

		    }
		}
	    }
	}

      channel[i]->type =
	(APU.DSP[APU_NON] & chs[i].m) ? SOUND_NOISE : SOUND_SAMPLE;
      channel[i]->volume_left = (short) (int8) APU.DSP[chs[i].o + APU_VOL_LEFT];
      channel[i]->volume_right = (short) (int8) APU.DSP[chs[i].o + APU_VOL_RIGHT];
      channel[i]->hertz = ((uint32) DSP_GET_PITCH (i) * SNES_SAMPLE_RATE) >> 12;
      channel[i]->frequency = (uint32) (((int64) (channel[i]->type == SOUND_NOISE ?
					     sd->noise_hertz : channel[i]->hertz) *
				    FIXED_POINT) / SNES_SAMPLE_RATE);

      channel[i]->count = 0;

      eVal = mix[i].eVal;
      if (eVal > D_ATTACK)
	eVal = D_ATTACK;
      else if (eVal < 0)
	eVal = 0;

      channel[i]->envx = eVal >> E_SHIFT;
      channel[i]->envxx = (channel[i]->envx << 24);	// #define ENVX_SHIFT 24
      channel[i]->left_vol_level = (channel[i]->envx * channel[i]->volume_left) / 128;
      channel[i]->right_vol_level = (channel[i]->envx * channel[i]->volume_right) / 128;

      switch (mix[i].eMode)
	{
	case E_ATT:
	  //Attack mode
	  //channel[i]->envx_target = MAX_ENVELOPE_HEIGHT;
	  channel[i]->envx_target = mix[i].eDest >> E_SHIFT;
	  break;

	case E_DECAY:
	  //Decay mode
	  //channel[i]->envx_target = (MAX_ENVELOPE_HEIGHT * ((APU.DSP[chs[i].o + APU_ADSR2] >> 5) + 1)) >> 3;
	  channel[i]->envx_target = mix[i].eDest >> E_SHIFT;
	  break;

	case E_SUST:
	  //Sustain mode
	  //channel[i]->envx_target = MAX_ENVELOPE_HEIGHT / 10;
	  channel[i]->envx_target = mix[i].eDest >> E_SHIFT;
	  break;

	default:
	  channel[i]->envx_target = 0;
	  break;
	}

      channel[i]->env_error = mix[i].eDec;
      if (so.playback_rate > 0)
	channel[i]->erate = mix[i].eRate * SNES_SAMPLE_RATE / so.playback_rate;
      else
	channel[i]->erate = 0;

      channel[i]->direction =
	mix[i].eRate == 0 ? 0 : ((mix[i].eMode & E_DIR) ? 1 : -1);

      if ((APU.DSP[chs[i].o + APU_ADSR1] & 0xF) < 0x0F)
	{
	  f = freqTab[((APU.DSP[chs[i].o + APU_ADSR1] & 0xF) << 1) + 1];
	  channel[i]->attack_rate = (unsigned long) (f << 1);
	}
      else
	channel[i]->attack_rate = (unsigned long) 1;

      // 1->1/10, eVal*255/256, T_EXP times.
      f = freqTab[((APU.DSP[chs[i].o + APU_ADSR1] & 0x70) >> 3) + 0x10];
      channel[i]->decay_rate =
	(unsigned long) (f * T_EXP / (SNES_SAMPLE_RATE / 1000));

      if ((APU.DSP[chs[i].o + APU_ADSR2] & 0x1F) > 0)
	{
	  f = freqTab[APU.DSP[chs[i].o + APU_ADSR2] & 0x1F];
	  channel[i]->sustain_rate =
	    (unsigned long) (f * T_EXP / (SNES_SAMPLE_RATE / 1000));
	}
      else
	channel[i]->sustain_rate = ~0;

      channel[i]->release_rate = 8;
      channel[i]->sustain_level = (APU.DSP[chs[i].o + APU_ADSR2] >> 5) + 1;
      channel[i]->interpolate = 0;

      if(mix[i].sIdx >= 0)
	{
	  memcpy (channel[i]->decoded, &mix[i].sBuf[mix[i].sIdx & 0x10],
		  sizeof (int16) * 16);
	  channel[i]->sample = mix[i].sBuf[mix[i].sIdx];
	  channel[i]->next_sample = mix[i].sBuf[mix[i].sIdx + 1];
	  channel[i]->sample_pointer = (uint32) (mix[i].sIdx & 0x0f);
	}
      else
	{
	  memset (channel[i]->decoded, 0, sizeof (int16) * 16);
	  channel[i]->sample = 0;
	  channel[i]->next_sample = 0;
	  channel[i]->sample_pointer = (uint32) (mix[i].sIdx & 0xff);
	}

      channel[i]->Filter = mix[i].bHdr;
      channel[i]->last_block = (channel[i]->Filter & 1) != 0;
      channel[i]->loop = (channel[i]->Filter & 2) != 0;

      channel[i]->previous[0] = mix[i].sP1;
      channel[i]->previous[1] = mix[i].sP2;
      channel[i]->previous16[0] = (signed short) mix[i].sP1;
      channel[i]->previous16[1] = (signed short) mix[i].sP2;
      channel[i]->block = (signed short *) &channel[i]->decoded[channel[i]->sample_pointer];
      channel[i]->sample_number = APU.DSP[chs[i].o + APU_SRCN];
      channel[i]->needs_decode = FALSE;
      channel[i]->block_pointer = (uint32) mix[i].bCur;

      channel[i]->echo_buf_ptr = NULL;

      channel[i]->chkMode = 0;

      channel[i]->dummy[0] = 0;
      channel[i]->dummy[1] = 0;
      channel[i]->dummy[2] = 0;
      channel[i]->dummy[3] = 0;
      channel[i]->dummy[4] = 0;
      channel[i]->dummy[5] = 0;
      channel[i]->dummy[6] = 0;
      channel[i]->dummy[7] = 0;
    }

  // Store 0xf3
  pAPURAM[0xf3] = APU.DSP[pAPURAM[0xf2] & 0x7f];
}

static const uint8 S9xtoSAEMode[10] = {
  E_IDLE,			// SOUND_SILENT
  E_ATT,			// SOUND_ATTACK
  E_DECAY,			// SOUND_DECAY
  E_SUST,			// SOUND_SUSTAIN
  E_REL,			// SOUND_RELEASE
  E_DIRECT,			// SOUND_GAIN
  E_INC,			// SOUND_INCREASE_LINEAR
  E_BENT,			// SOUND_INCREASE_BENT_LINE
  E_DEC,			// SOUND_DECREASE_LINEAR
  E_EXP				// SOUND_DECREASE_EXPONENTIAL
};

// Restore DSP's State of Operation
void
RestoreAPUDSP ()
{
  int i;

  voiceKon = APU.KeyedChannels;
  voiceKoff = APU.DSP[APU_KOFF];

  for (i = 0; i < 8; i++)
    {
      mix[i].bCur = (uint16) (channel[i]->block_pointer & 0xffff);
      mix[i].bHdr = (uint8) channel[i]->Filter;
      memset (&mix[i].sBuf[0x10], 0, sizeof (int16) * 16);
      memcpy (mix[i].sBuf, channel[i]->decoded, sizeof (int16) * 16);
      mix[i].sP1 = (int32) channel[i]->previous16[0];
      mix[i].sP2 = (int32) channel[i]->previous16[1];

      if(channel[i]->sample_pointer <= 255 && channel[i]->sample_pointer >= 247)
	mix[i].sIdx = (int8)(channel[i]->sample_pointer & 0xff);
      else
	mix[i].sIdx = (int8)(channel[i]->sample_pointer & 0x0f);

      uint8 mFlg = ((mix[i].mFlg & MFLG_MUTE) |
		    ((channel[i]->last_block && !channel[i]->loop) ? MFLG_END : 0) |
		    (channel[i]->state == SOUND_SILENT ? MFLG_OFF : 0) |
		    (channel[i]->mode == MODE_RELEASE ? MFLG_KOFF : 0));

      if (mFlg & MFLG_OFF)
	mFlg &= ~(MFLG_KOFF | MFLG_END);

      mix[i].mFlg = mFlg;

      InitReg (chs[i].o + APU_VOL_LEFT, APU.DSP[chs[i].o + APU_VOL_LEFT]);
      InitReg (chs[i].o + APU_VOL_RIGHT, APU.DSP[chs[i].o + APU_VOL_RIGHT]);
      InitReg (chs[i].o + APU_P_LOW, APU.DSP[chs[i].o + APU_P_LOW]);

      mix[i].eMode = S9xtoSAEMode[channel[i]->state];

      // Default to release mode settings
      mix[i].eRIdx = 31;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      mix[i].eDec = 0;
      mix[i].eAdj = A_KOF;
      mix[i].eDest = D_MIN;
      mix[i].eVal = channel[i]->envx << E_SHIFT;
      InitReg (chs[i].o + APU_ADSR1, APU.DSP[chs[i].o + APU_ADSR1]);

      mix[i].mOut = (int32) (int8) APU.DSP[chs[i].o + APU_OUTX] << 8;
      mix[i].mDec = 0;
    }
  // Setup global paramaters
  InitReg (APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
  InitReg (APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
  InitReg (APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
  InitReg (APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

  // Don't "APU_SOFT_RESET"
  uint8 flg = APU.DSP[APU_FLG];
  RFlg ((APU_FLG & 0x70) >> 4, flg & ~APU_SOFT_RESET);
  APU.DSP[APU_FLG] = flg;

  InitReg (APU_ESA, APU.DSP[APU_ESA]);
  InitReg (APU_EFB, APU.DSP[APU_EFB]);
  InitReg (APU_EDL, APU.DSP[APU_EDL]);

  for (i = 0; i < 8; i++)
    InitReg (chs[i].o + APU_C0, APU.DSP[chs[i].o + APU_C0]);
}

void
RestoreSPCFile ()
{
  int i;

  for (i = 0; i < 8; i++)
    {
      mix[i].eVal = APU.DSP[chs[i].o + APU_ENVX] << E_SHIFT;
      //mix[i].mOut = (int32)(int8)APU.DSP[chs[i].o + APU_OUTX] << 8;
    }

  // Setup global paramaters
  InitReg (APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
  InitReg (APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
  InitReg (APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
  InitReg (APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

  // Don't "APU_SOFT_RESET"
  uint8 flg = APU.DSP[APU_FLG];
  RFlg ((APU_FLG & 0x70) >> 4, flg & ~APU_SOFT_RESET);
  APU.DSP[APU_FLG] = flg;

  InitReg (APU_ESA, APU.DSP[APU_ESA]);
  InitReg (APU_EFB, APU.DSP[APU_EFB]);
  InitReg (APU_EDL, APU.DSP[APU_EDL]);

  for (i = 0; i < 8; i++)
    InitReg (chs[i].o + APU_C0, APU.DSP[chs[i].o + APU_C0]);

  // Start Voices
  InitReg (APU_KON, APU.DSP[APU_KON]);
}

// DSP Pre-Amplification Level
void
SetAPUDSPAmp (int32 amp)
{
  volAdj = amp;

  // Update global volumes
  InitReg (APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
  InitReg (APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
  InitReg (APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
  InitReg (APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);
}

static inline void
ChgSus (int i)
{
  mix[i].eRIdx = APU.DSP[chs[i].o + APU_ADSR2] & 0x1f;
  if (rateTab[mix[i].eRIdx] == 0 || mix[i].eVal <= D_MIN)
    {
      mix[i].eMode = E_IDLE | E_SUST;
      mix[i].eDec = 0;
    }
  else
    mix[i].eMode = E_SUST;
  mix[i].eRate = rateTab[mix[i].eRIdx];
  mix[i].eAdj = A_EXP;
  mix[i].eDest = D_MIN;
}

static inline void
ChgDec (int i)
{
  uint8 sl = APU.DSP[chs[i].o + APU_ADSR2] >> 5;
  int32 eDest = (int32) (sl + 1) * D_EXP;
  if (sl == 7 || mix[i].eVal <= eDest)
    {
      ChgSus (i);
      return;
    }
  mix[i].eRIdx = ((APU.DSP[chs[i].o + APU_ADSR1] & 0x70) >> 3) + 0x10;
  mix[i].eRate = rateTab[mix[i].eRIdx];
  mix[i].eAdj = A_EXP;
  mix[i].eDest = eDest;
  mix[i].eMode = E_DECAY;
}

static inline void
ChgAtt (int i)
{
  uint8 ar = APU.DSP[chs[i].o + APU_ADSR1] & 0x0f;

  if (mix[i].eVal >= D_ATTACK)
    {
      ChgDec (i);
      return;
    }
  mix[i].eRIdx = (ar << 1) + 1;
  mix[i].eRate = rateTab[mix[i].eRIdx];
  mix[i].eAdj = ar == 0x0f ? A_NOATT : A_LIN;
  mix[i].eDest = D_ATTACK;
  mix[i].eMode = E_ATT;
}

static inline void
ChgADSR (int i)
{
  switch (mix[i].eMode)
    {
    case E_ATT:
      ChgAtt (i);
      break;

    case E_DECAY:
      ChgDec (i);
      break;

    case E_SUST:
    case (E_IDLE | E_SUST):
      if (mix[i].eVal > D_MIN)
	ChgSus (i);
      break;
    }
}

static inline void
ChgGain (int i)
{
  // Is gain direct?
  if (!(APU.DSP[chs[i].o + APU_GAIN] & 0x80))
    {
      mix[i].eMode = E_DIRECT;
      mix[i].eRIdx = 0;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      mix[i].eAdj = A_EXP;
      mix[i].eVal = mix[i].eDest =
	(int32) (APU.DSP[chs[i].o + APU_GAIN] & 0x7f) * A_GAIN;
      return;
    }

  switch (APU.DSP[chs[i].o + APU_GAIN] & 0x60)
    {
    case 0x00:
      // GainDec
      mix[i].eRIdx = APU.DSP[chs[i].o + APU_GAIN] & 0x1f;
      if (rateTab[mix[i].eRIdx] == 0 || mix[i].eVal <= D_MIN)
	{
	  mix[i].eMode = E_IDLE | E_DEC;
	  mix[i].eDec = 0;
	}
      else
	mix[i].eMode = E_DEC;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      mix[i].eAdj = A_LIN;
      mix[i].eDest = D_MIN;
      break;

    case 0x20:
      // GainExp
      mix[i].eRIdx = APU.DSP[chs[i].o + APU_GAIN] & 0x1f;
      if (rateTab[mix[i].eRIdx] == 0 || mix[i].eVal <= D_MIN)
	{
	  mix[i].eMode = E_IDLE | E_EXP;
	  mix[i].eDec = 0;
	}
      else
	mix[i].eMode = E_EXP;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      mix[i].eAdj = A_EXP;
      mix[i].eDest = D_MIN;
      break;

    case 0x40:
      // GainInc
      mix[i].eRIdx = APU.DSP[chs[i].o + APU_GAIN] & 0x1f;
      if (rateTab[mix[i].eRIdx] == 0 || mix[i].eVal >= D_ATTACK)
	{
	  mix[i].eMode = E_IDLE | E_INC;
	  mix[i].eDec = 0;
	}
      else
	mix[i].eMode = E_INC;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      mix[i].eAdj = A_LIN;
      mix[i].eDest = D_ATTACK;
      break;

    case 0x60:
      // GainBent
      mix[i].eRIdx = APU.DSP[chs[i].o + APU_GAIN] & 0x1f;
      if (rateTab[mix[i].eRIdx] == 0 || mix[i].eVal >= D_ATTACK)
	{
	  mix[i].eMode = E_IDLE | E_BENT;
	  mix[i].eDec = 0;
	}
      else
	mix[i].eMode = E_BENT;
      mix[i].eRate = rateTab[mix[i].eRIdx];
      if (mix[i].eVal < D_BENT)
	{
	  mix[i].eAdj = A_LIN;
	  mix[i].eDest = D_BENT;
	}
      else
	{
	  mix[i].eAdj = A_BENT;
	  mix[i].eDest = D_ATTACK;
	}
      break;
    }
}

// Start Envelope
//  Called when a voice is keyed on to set up the internal data to begin
//  envelope modification based on the values in ADSR/Gain.
static inline void
StartEnv (int i)
{
  if ((mix[i].mFlg & (MFLG_OFF | MFLG_END)) || (mix[i].eMode & E_IDLE))
    mix[i].eDec = 0;

  if (APU.DSP[chs[i].o + APU_ADSR1] & 0x80)
    ChgAtt (i);
  else
    ChgGain (i);
}

static inline void
UpdateEnv (int i)
{
  // Is the adjustment exponential? (mode 1, 9, or D)
  // E_EXP, E_SUST, E_DECAY
  if (mix[i].eMode & E_TYPE)
    {
      if (mix[i].eVal > mix[i].eDest)
	{
	  //mix[i].eVal = (mix[i].eVal * 255) / 256;
	  mix[i].eVal--;
	  mix[i].eVal -= mix[i].eVal >> 8;
	  if (mix[i].eVal > mix[i].eDest)
	    return;
	  if(mix[i].eVal < mix[i].eDest)
	    mix[i].eVal = mix[i].eDest;
	}
    }
  // Linear increase (mode 2, 6, or A)
  // E_INC, E_BENT, E_ATT
  else if (mix[i].eMode & E_DIR)
    {
      if (mix[i].eVal < mix[i].eDest)
	{
	  mix[i].eVal += mix[i].eAdj;
	  if (mix[i].eVal < mix[i].eDest)
	    return;
	  if(mix[i].eVal > mix[i].eDest)
	    mix[i].eVal = mix[i].eDest;
	}
    }
  // Linear decrease (mode 0 or 8)
  // E_DEC, E_REL
  else
    {
      if (mix[i].eVal > mix[i].eDest)
	{
	  if (mix[i].eVal >= mix[i].eAdj)
	    mix[i].eVal -= mix[i].eAdj;
	  else
	    mix[i].eVal = mix[i].eDest;
	  if (mix[i].eVal > mix[i].eDest)
	    return;
	}
      if (mix[i].eMode == E_REL)
	{			// Is voice in release mode?
	  mix[i].mFlg |= MFLG_OFF;
	  mix[i].mFlg &= ~MFLG_KOFF;
	}
    }

  // Is envelope in ADSR mode?
  // E_REL, E_SUST, E_ATT, E_DECAY
  if (mix[i].eMode & E_ADSR)
    {
      // Change adjustment mode
      switch (mix[i].eMode)
	{
	case E_ATT:
	  if (mix[i].eVal >= mix[i].eDest)
	    ChgDec (i);
	  return;

	case E_DECAY:
	  if (mix[i].eVal <= mix[i].eDest)
	    ChgSus (i);
	  return;

	case E_SUST:
	  break;

	case E_REL:
	  break;
	}
    }
  // Is gain in bent line mode? No, it's constant
  else if (mix[i].eMode & E_DEST)
    {
      if (mix[i].eDest == D_BENT)
	{
	  mix[i].eAdj = A_BENT;
	  mix[i].eDest = D_ATTACK;
	  return;
	}
    }
  mix[i].eMode |= E_IDLE;
  mix[i].eVal = mix[i].eDest;	//clip eVal
  mix[i].eDec = 0;
}

static inline void
CalcEnv (int i)
{
  if (!(mix[i].eMode & E_IDLE))
    {
      uint32 e = (mix[i].eDec += mix[i].eRate) >> FIXED_POINT_SHIFT;
      mix[i].eDec &= FIXED_POINT_REMAINDER;
      for (; e; e--)
	{
	  UpdateEnv (i);
	  if (mix[i].eMode & E_IDLE)
	    break;
	}
    }
}

static inline void ChgSilence(int ch)
{
  mix[ch].eMode = E_DIRECT;
  mix[ch].eRIdx = 0;
  mix[ch].eRate = rateTab[mix[ch].eRIdx];
  mix[ch].eAdj = A_EXP;
  mix[ch].eVal = mix[ch].eDest = D_MIN;
  mix[ch].mFlg |= MFLG_OFF;
  mix[ch].mFlg &= (MFLG_MUTE | MFLG_OFF);
}


static inline void ChgRelease(int ch)
{
  mix[ch].eRIdx = 0x1f;	// 1->0 8ms
  mix[ch].eRate = rateTab[mix[ch].eRIdx];
  mix[ch].eAdj = A_KOF;
  mix[ch].eDest = D_MIN;
  mix[ch].eMode = E_REL;
  mix[ch].mFlg |= MFLG_KOFF;	// Flag voice as keying off
}

static inline void
ProcessSrc (int i)
{
  // Waveform Resizing
  int8 sc = (int8) ((mix[i].mDec += mix[i].mRate) >> FIXED_POINT_SHIFT);
  mix[i].mDec &= FIXED_POINT_REMAINDER;
  int8 sIdx = mix[i].sIdx + sc;

  if(sIdx < 0)
    {
      mix[i].sIdx = sIdx;
      return;
    }

  sIdx &= 0x1f;

  bool8 firstblk = FALSE;
  if(mix[i].sIdx < 0)
    firstblk = TRUE;

  // It needs next 16 samples?
  if (firstblk || ((mix[i].sIdx ^ sIdx) & 0x10))
    {
      mix[i].sIdx = sIdx;
      UpdateSrc (i, firstblk);
    }
  else
    mix[i].sIdx = sIdx;
}

// DSP Data Port
void
APUDSPIn (uint8 address, uint8 data)
{
  // Writes to 80-FFh have no effect (reads are mirrored from lower mem)
  if (address & 0x80)
    return;

  switch (address)
    {
    case APU_KON:
    case APU_KOFF:
      InitReg (address, data);
      break;

    case APU_FLG:
      RFlg ((address & 0x70) >> 4, data);
      break;

    case APU_ENDX:
      REndX ((address & 0x70) >> 4, data);
      break;

    default:
      if (data != APU.DSP[address])
	InitReg (address, data);
      break;
    }
}

/***********************************************************************
  DSP Register Handlers
***********************************************************************/

// Channel volume
static void
RVolL (int i, uint8 val)
{
  mix[i].mChnL = (int32) (int8) val;
}

static void
RVolLM (int i, uint8 val)
{
  mix[i].mChnL = (int32) ABS ((int8) val);
}

static void
RVolR (int i, uint8 val)
{
  mix[i].mChnR = (int32) (int8) val;
}

static void
RVolRM (int i, uint8 val)
{
  mix[i].mChnR = (int32) ABS ((int8) val);
}

// Pitch
static void
RPitch (int i, uint8)
{
  mix[i].mOrgP = DSP_GET_PITCH (i);
  uint64 r = (uint64) mix[i].mOrgP * pitchAdj;
  mix[i].mOrgRate =
    (uint32) ((r >> FIXED_POINT_SHIFT) +
	      ((r & FIXED_POINT_REMAINDER) ? 1 : 0));
}

static void
RSrcn (int, uint8)
{
  /* Do nothing */
}

// Envelope
static void
RADSR1 (int i, uint8 val)
{
  // MFLG_OFF was checked in InitReg()
  if (mix[i].mFlg & MFLG_KOFF)
    return;

  if (APU.DSP[chs[i].o + APU_ADSR1] & 0x80)
    ChgADSR (i);
  else
    ChgGain (i);
}

static void
RADSR2 (int i, uint8 val)
{
  // MFLG_OFF was checked in InitReg()
  if ((mix[i].mFlg & MFLG_KOFF) || !(APU.DSP[chs[i].o + APU_ADSR1] & 0x80))
    return;
  ChgADSR (i);
}

static void
RGain (int i, uint8)
{
  // MFLG_OFF was checked in InitReg()
  if ((mix[i].mFlg & MFLG_KOFF) || (APU.DSP[chs[i].o + APU_ADSR1] & 0x80))
    return;
  ChgGain (i);
}

// Main volumes
static void
RMVolL (int, uint8 val)
{
  if (Settings.DisableMasterVolume)
    val = 127;
  volMainL = ((int32) (int8) val * volAdj) >> FIXED_POINT_SHIFT;
}

static void
RMVolLM (int, uint8 val)
{
  if (Settings.DisableMasterVolume)
    val = 127;
  // Multiply by 1/2 sqrt 2 to help average
  volMainL =
    (int32) (((int64) ABS ((int8) val) * 46341 *
	      volAdj) >> (FIXED_POINT_SHIFT + 16));
}

static void
RMVolR (int, uint8 val)
{
  if (Settings.DisableMasterVolume)
    val = 127;
  volMainR = ((int32) (int8) val * volAdj) >> FIXED_POINT_SHIFT;
}

static void
RMVolRM (int, uint8 val)
{
  if (Settings.DisableMasterVolume)
    val = 127;
  // Multiply by 1/2 sqrt 2 to help average
  volMainR =
    (int32) (((int64) ABS ((int8) val) * 46341 *
	      volAdj) >> (FIXED_POINT_SHIFT + 16));
}

static void
REVolL (int, uint8 val)
{
  volEchoL = ((int32) (int8) val * volAdj) >> FIXED_POINT_SHIFT;
}

static void
REVolLM (int, uint8 val)
{
  // Multiply by 1/2 sqrt 2 to help average
  volEchoL =
    (int32) (((int64) ABS ((int8) val) * 46341 *
	      volAdj) >> (FIXED_POINT_SHIFT + 16));
}

static void
REVolR (int, uint8 val)
{
  volEchoR = ((int32) (int8) val * volAdj) >> FIXED_POINT_SHIFT;
}

static void
REVolRM (int, uint8 val)
{
  // Multiply by 1/2 sqrt 2 to help average
  volEchoR =
    (int32) (((int64) ABS ((int8) val) * 46341 *
	      volAdj) >> (FIXED_POINT_SHIFT + 16));
}

// Echo settings
static void
REFB (int, uint8 val)
{
  echoFB = (int32) (int8) val;
}

static void
REFBM (int, uint8 val)
{
  echoFB = (int32) (int8) val;
}

static void
REDl (int, uint8 val)
{
  val &= 0x0f;
  if (val == 0)
    echoDel = 2;
  else
    echoDel = ((uint32) (val << 4) * 32) << 1;
}

static void
RFCI (int i, uint8 val)
{
  FilterTaps[i] = (int8) val;
}

// Key On/Off
static void
RKOn (int idx, uint8 val)
{
  if (spc_is_dumping)
    {
      if (val & ~spc_is_dumping_temp)
	{
#ifdef __UOSNES__
#ifdef __WIN32__
	  extern void uowSPCDumpMessages (bool8 err);
	  uowSPCDumpMessages (S9xSPCDump (S9xGetFilenameInc (".spc", SPC_DIR)));
#endif
#else
	  S9xSPCDump (S9xGetFilenameInc (".spc", SPC_DIR));
#endif
	  spc_is_dumping = 0;
	}
    }
  spc_is_dumping_temp = val;

  voiceKon = val;
}

static void
RKOf (int, uint8 val)
{
  voiceKoff = val;
}

static void
RDir (int, uint8)
{
}

static void
RPMOn (int, uint8)
{
}

static void
RNOn (int, uint8)
{
}

static void
RFlg (int, uint8 al)
{
  // Disable echo
  disEcho = (disEcho & ~APU_ECHO_DISABLED) | (al & APU_ECHO_DISABLED);
  APU.DSP[APU_FLG] = al;
  SetNoiseHertz ();
}

static void
RESA (int, uint8 val)
{
  echoStart = (val * 64) << 1;
}

static void
REndX (int, uint8)
{
  APU.DSP[APU_ENDX] = 0;
}

// Null register
static void
RNull (int, uint8)
{
  /* Do nothing */
}

// Noise Generator
//  Generates white noise samples (into noise)
static inline void
NoiseGen ()
{
  noise_ctr += noise_rate;

  if(noise_ctr >= 0x7800)
    {
      noise_ctr -= 0x7800;
      int feedback = (noise << 13) ^ (noise << 14);
      noise = (feedback & 0x4000) ^ (noise >> 1);
    }
}

// Pitch Modulation
//
// Changes the pitch based on the output of the previous voice:
//
//  P' = (P * (OUTX + 32768)) >> 15
//
// Pitch modulation in the SNES uses the full 16-bit sample value, not the 8-bit value in OUTX as
// previously believed.
static inline void
PitchMod (int i)
{
  int32 t = ((mix[i - 1].mOut + 32768) * mix[i].mOrgP) >> 15;

  if (t > 0x3FFF)
    t = 0x3FFF;
  else if (t < 0)
    t = 0;

  uint64 r = (uint64) t * pitchAdj;
  mix[i].mRate =
    (uint32) ((r >> FIXED_POINT_SHIFT) +
	      ((r & FIXED_POINT_REMAINDER) ? 1 : 0));
}

static inline int32
GetCurSample (int ch, int InterpolatedSound)
{
  if(mix[ch].sIdx < 0)
    return 0;

  switch (InterpolatedSound)
    {
    case INT_LINEAR:
      {
	// Linear Interpolation
	int16 ps = mix[ch].sBuf[(mix[ch].sIdx - 1) & 0x1f];
	return (int32) ((((((int32) mix[ch].sBuf[mix[ch].sIdx] - ps) >> 1) *
			  (int32) mix[ch].mDec) >> 15) + ps);
      }

    case INT_CUBIC:
      {
	// Cubic Interpolation
	int mDec = (mix[ch].mDec >> 8) << 2;
	int32 cursample = (int32) (int16) ((((int32) mix[ch].sBuf[(mix[ch].sIdx - 3) & 0x1f] * cubicTab[mDec + 0]) >> 16)	//d0
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 2) & 0x1f] * cubicTab[mDec + 1]) >> 16)	//d1
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 1) & 0x1f] * cubicTab[mDec + 2]) >> 16)	//d2
					   + (((int32) mix[ch].sBuf[mix[ch].sIdx] * cubicTab[mDec + 3]) >> 16));	//d3
	cursample += cursample;
	return cursample;
      }

    case INT_GAUSS:
      {
	// Gaussian Interpolation
	int mDec = (mix[ch].mDec >> 8) << 2;
	int32 cursample = (int32) (int16) ((((int32) mix[ch].sBuf[(mix[ch].sIdx - 3) & 0x1f] * gaussTab[mDec + 0]) >> 16)	//d0
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 2) & 0x1f] * gaussTab[mDec + 1]) >> 16)	//d1
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 1) & 0x1f] * gaussTab[mDec + 2]) >> 16)	//d2
					   + (((int32) mix[ch].sBuf[mix[ch].sIdx] * gaussTab[mDec + 3]) >> 16));	//d3
	cursample += cursample;
	return cursample;
      }

    case INT_SINC:
      {
	//ﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄﾄ
	//	8-point Sinc Interpolation
	//
	//	Ideally the sinc function sould be used as low-pass filter when the playback rate exceeds the
	//	output rate.
	int mDec = (mix[ch].mDec >> 5) & ~7;
	int32 cursample = (int32) (int16) ((((int32) mix[ch].sBuf[(mix[ch].sIdx - 7) & 0x1f] * sincTab[mDec + 0]) >> 16)	//d0
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 6) & 0x1f] * sincTab[mDec + 1]) >> 16)	//d1
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 5) & 0x1f] * sincTab[mDec + 2]) >> 16)	//d2
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 4) & 0x1f] * sincTab[mDec + 3]) >> 16)	//d3
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 3) & 0x1f] * sincTab[mDec + 4]) >> 16)	//d4
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 2) & 0x1f] * sincTab[mDec + 5]) >> 16)	//d5
					   + (((int32) mix[ch].sBuf[(mix[ch].sIdx - 1) & 0x1f] * sincTab[mDec + 6]) >> 16)	//d6
					   + (((int32) mix[ch].sBuf[mix[ch].sIdx] * sincTab[mDec + 7]) >> 16));	//d7
	cursample += cursample;
	return cursample;
      }
    }
  //case INT_NONE:
  return (int32) mix[ch].sBuf[mix[ch].sIdx];
}

static inline int32
ProcessSample (int ch)
{
  // Pitch Modulation
  if (APU.DSP[APU_PMON] & ~APU.DSP[APU_NON] & ~chs[0].m & chs[ch].m)
    PitchMod (ch);
  else
    mix[ch].mRate = mix[ch].mOrgRate;

  // Waveform Resizing
  ProcessSrc (ch);

  int32 eVal = mix[ch].eVal;

  if (eVal > D_ATTACK)
    eVal = D_ATTACK;
  else if (eVal < D_MIN)
    eVal = D_MIN;

  int32 cursample;

  if (!(APU.DSP[APU_NON] & chs[ch].m))
    {
      mix[ch].mOut = ((GetCurSample (ch, INT_GAUSS) * eVal) >> (E_SHIFT + 7)) & ~1;
      cursample = ((GetCurSample (ch, Settings.InterpolatedSound) * eVal) >> (E_SHIFT + 7)) & ~1;
    }
  else
    {
      mix[ch].mOut = cursample = (((int)((int16)(noise << 1)) * eVal) >> (E_SHIFT + 7)) & ~1;
    }

  return cursample;
}

// Finite Impulse Response Echo Filter
// 
// Filters the echo using an eight tap FIR filter:
// 
//         7
//        ---
//    x = \   c  * s
//        /    n    n
//        ---
//        n=0
// 
//    x = output sample
//    c = filter coefficient (-.7)
//    s = unfiltered sample
//    n = 0 is the oldest sample and 7 is the most recent
// 
// FIR filters are based on the sample rate.  This was fine in the SNES, because the sample rate was
// always 32kHz, but in the case of an emulator the sample rate can change.  So measures have to be
// taken to ensure the filter will have the same effect, regardless of the output sample rate.
// 
// To overcome this problem, I figured each tap of the filter is applied every 31250ns.  So the
// solution is to calculate when 31250ns have gone by, and use the sample at that point.  Of course
// this method really only works if the output rate is a multiple of 32k.  In order to get accurate
// results, some sort of interpolation method needs to be introduced.  I went the cheap route and used
// linear interpolation.
static inline void
FIRFilter (int32 &l, int32 &r)
{
  l = 0;
  r = 0;

  // Reset decimal overflow, so filtering is consistant
  uint8 fi = firCur;

  for (int i = 0; i < 8; i++)
    {
      l += (Loop[fi + 0] * FilterTaps[i]) >> 6;
      r += (Loop[fi + 1] * FilterTaps[i]) >> 6;
      fi = (fi - 2) & (FIRBUF - 2);
    }
}

static inline void
ECHOFilter (int32 &l, int32 &r, int32 ll, int32 rr)
{
  if(!Settings.SPCPlaying)
    {
      Loop[firCur + 0] = (int16)DSP_GET_WORD((echoStart + echoCur + 0) << 1) >> 1;
      Loop[firCur + 1] = (int16)DSP_GET_WORD((echoStart + echoCur + 1) << 1) >> 1;
    }
  else
    {
      Loop[firCur + 0] = (int16)DSP_GET_ECHO((echoStart + echoCur + 0) << 1) >> 1;
      Loop[firCur + 1] = (int16)DSP_GET_ECHO((echoStart + echoCur + 1) << 1) >> 1;
    }

  // Filter echo
  FIRFilter (l, r);

  firCur = (firCur + 2) & (FIRBUF - 2);

  // Add feedback
  if (!(APU.DSP[APU_FLG] & APU_ECHO_DISABLED))
    {
      if(!Settings.SPCPlaying)
	{
	  DSP_SET_WORD((echoStart + echoCur + 0) << 1, sclamp16(ll + ((l * echoFB) >> 7)) & ~1);
	  DSP_SET_WORD((echoStart + echoCur + 1) << 1, sclamp16(rr + ((r * echoFB) >> 7)) & ~1);
	}
      else
	{
	  DSP_SET_ECHO((echoStart + echoCur + 0) << 1, sclamp16(ll + ((l * echoFB) >> 7)) & ~1);
	  DSP_SET_ECHO((echoStart + echoCur + 1) << 1, sclamp16(rr + ((r * echoFB) >> 7)) & ~1);
	}
    }

  echoCur = (echoCur + 2) % echoDel;
}

void ClearSamples (void)
{
  resampler->clear();
}

void SetBufferSize (int ms)
{
  resampler->resize(65 * ms * 2);
}

//**************************************************************************************************
// Emulate DSP
//
// Emulates the DSP of the SNES
//
void
EmuAPUDSP (uint8 *pBuf, int32 num)
{
  static int16 s[SYNCBUF*2];

  if(resampler->avail() >= (num << 1) + lag)
    {
      resampler->read(s, (num << 1));
      if(lag != 0)
	lag = 0;
    }
  else
    {
      memset(s, 0, num * sizeof(int16) * 2);
      if(lag == 0)
	lag = (num << 1) - resampler->avail();
    }

  for (int cnt = 0; cnt < num; cnt++, pBuf += (dspSize * dspChn))
    {
      int16 l, r;

      // Render sound buffer
      switch (dspChn)
	{
	case 1:// mono
	  switch (dspSize)
	    {
	    case 1:// 8bit
	      l = (int16)(((((int32)s[(cnt << 1) + 0] + s[(cnt << 1) + 1]) * 181) >> 16) + 128);

	      if (l < 0)
		l = 0;
	      else if (l > 255)
		l = 255;

	      pBuf[0] = (uint8) l;
	      break;

	    case 2:// 16bit
	      l = (int16)((((int32)s[(cnt << 1) + 0] + s[(cnt << 1) + 1]) * 181) >> 8);
	      ((int16 *) pBuf)[0] =  l;
	      break;
	    }
	  break;

	case 2:// stereo
	  switch (dspSize)
	    {
	    case 1:// 8bit
	      l = (int16)(((int32)s[(cnt << 1) + 0] + 32768) >> 8);	// Unsign sample
	      r = (int16)(((int32)s[(cnt << 1) + 1] + 32768) >> 8);	// Unsign sample

	      if (l < 0)
		l = 0;
	      else if (l > 255)
		l = 255;

	      if (r < 0)
		r = 0;
	      else if (r > 255)
		r = 255;

	      pBuf[0] = (uint8) l;
	      pBuf[1] = (uint8) r;
	      break;

	    case 2:// 16bit
	      ((int16 *) pBuf)[0] = s[(cnt << 1) + 0];
	      ((int16 *) pBuf)[1] = s[(cnt << 1) + 1];
	      break;
	    }
	  break;
	}
    }
}

void smpdsp_run()
{
  if((dsp_counter++ & 1) == 0)
    {
      for(int ch = 0; ch < 8; ch++)
	{
	  if(APU.DSP[APU_FLG] & APU_SOFT_RESET)
	    {
	      if(!(mix[ch].mFlg & MFLG_OFF))
		ChgSilence(ch);
	    }
	  if(voiceKoff & chs[ch].m)
	    {
	      if(!(mix[ch].mFlg & (MFLG_OFF | MFLG_KOFF)))
		{
		  ChgRelease(ch);
		}
	    }
	  if(voiceKon & chs[ch].m)
	    {
	      // Set channel volume
	      (*dspRegs[chs[ch].o + APU_VOL_LEFT]) (ch, APU.DSP[chs[ch].o + APU_VOL_LEFT]);
	      (*dspRegs[chs[ch].o + APU_VOL_RIGHT]) (ch, APU.DSP[chs[ch].o + APU_VOL_RIGHT]);

	      // Set pitch
	      RPitch (ch, APU.DSP[chs[ch].o + APU_P_LOW]);

	      StartEnv (ch);

	      // Start waveform decompression
	      mix[ch].mFlg &= ~(MFLG_KOFF | MFLG_OFF);	// Reset flags
	      StartSrc (ch);
	    }
	}
      APU.DSP[APU_ENDX] &= ~voiceKon;
      voiceKon = 0;
    }

  // Generate Noise
  NoiseGen ();

  // Erase current samples
  int32 MixSampleL = 0, MixSampleR = 0;
  int32 EchoSampleL = 0, EchoSampleR = 0;

  for (int ch = 0; ch < 8; ch++)
    {
      int32 cursample = ProcessSample (ch);

      // Is the current voice active?
      if (!(mix[ch].mFlg & MFLG_OFF))
	{
	  // Mixing
	  if (cursample != 0 && !(mix[ch].mFlg & MFLG_MUTE))
	    {
	      MixSampleL += (cursample * mix[ch].mChnL) >> 7;// Add to master samples
	      MixSampleR += (cursample * mix[ch].mChnR) >> 7;
	      if (APU.DSP[APU_EON] & chs[ch].m)// Is echo on?
		{
		  EchoSampleL += (cursample * mix[ch].mChnL) >> 7;
		  EchoSampleR += (cursample * mix[ch].mChnR) >> 7;
		}
	    }

	  // Envelope Calculation
	  CalcEnv (ch);
	  
	}//if(!(mix[i].mFlg & MFLG_OFF))
      else
	mix[ch].mOut = 0;

      int32 eVal = mix[ch].eVal >> E_SHIFT;

      if (eVal > 0x7f)
	eVal = 0x7f;
      else if (eVal < 0)
	eVal = 0;

      APU.DSP[chs[ch].o + APU_ENVX] = eVal;
      APU.DSP[chs[ch].o + APU_OUTX] = mix[ch].mOut >> 8;
    }//for(i = 0; i < 8; i++)

  int32 l, r;
  ECHOFilter (l, r, EchoSampleL, EchoSampleR);

  if (!(APU.DSP[APU_FLG] & APU_MUTE))// Is the DSP not muted?
    {
      // Multiply samples by main volume
      MixSampleL = (MixSampleL * volMainL) >> 7;
      MixSampleR = (MixSampleR * volMainR) >> 7;

      if (!(disEcho & DSP_NOECHO))
	{
	  MixSampleL += (l * volEchoL) >> 7;
	  MixSampleR += (r * volEchoR) >> 7;
	}
    }
  else
    {//if(APU.DSP[APU_FLG] & APU_MUTE)
      MixSampleL = 0;
      MixSampleR = 0;
    }

  WriteSamples(MixSampleL, MixSampleR);
}

void WriteSamples0(int32 l, int32 r)
{
  int16 s[2];
  s[0] = sclamp16(l);
  s[1] = sclamp16(r);
  resampler->push(s, 2);
}

//Decompress Sound Source
//
//Decompresses a 9-byte bit-rate reduced block into 16 16-bit samples
//
//In:
//  blk_hdr     = Block header
//  xsample_blk = Sample Block
//  output_buf  = Output buffer
//  smp_1       = Last sample of previous block
//  smp_2       = Next to last sample
//
//Out:
//  xsample_blk = Next Block
//  smp_1       = Last sample
//  smp_2       = Next to last sample
static inline int UnpckSrcFilter(int f, int p1, int p2)
{
  switch (f)
    {
    case 1:
      // Method1 - [Delta]+[Smp-1](15/16)
      return ((p1 >> 1) + ((-p1) >> 5));
    case 2:
      // Method2 - [Delta]+[Smp-1](61/32)-[Smp-2](15/16)
      return (p1 - (p2 >> 1) + (p2 >> 5) + ((p1 * -3) >> 6));
    case 3:
      // Method3 - [Delta]+[Smp-1](115/64)-[Smp-2](13/16)
      return (p1 - (p2 >> 1) + ((p1 * -13) >> 7) + (((p2 >> 1) * 3) >> 4));
    }
  // Method0 - [Smp]
  return 0;
}

static void
UnpckSrc (uint8 blk_hdr,
	  uint16 &xsample_blk,
	  int16 *output_buf, int32 &smp_1, int32 &smp_2)
{
  uint8 *sample_blk = DSP_GET_SRC_P (xsample_blk);
  xsample_blk += 9;

  //skip header
  sample_blk++;
  const int *BRR_row = &brrTab[(blk_hdr & 0xf0)];

  int f = (blk_hdr & 0x0c) >> 2;

  for (int i = 0; i < 8; i++)
    {
      smp_2 = BRR_row[sample_blk[0] >> 4] + UnpckSrcFilter(f, smp_1, smp_2);
      smp_2 = (int16)(sclamp16(smp_2) << 1);
      output_buf[0] = smp_2;

      smp_1 = BRR_row[sample_blk[0] & 0x0f] + UnpckSrcFilter(f, smp_2, smp_1);
      smp_1 = (int16)(sclamp16(smp_1) << 1);
      output_buf[1] = smp_1;

      sample_blk++;
      output_buf += 2;
    }
}

